package PureMessage::Enduser::Response;
use strict;
use Carp;
use PureMessage::Manager::Lang qw(%text &text &mtext &ntext &lang_text_utf8 &lang_ntext_utf8);
use PureMessage::Manager::Response;
use PureMessage::Enduser::Config qw(%config);
use PureMessage::Enduser::UI::Config qw(%enduser_ui_config);
use PureMessage::Enduser::Features;
use PureMessage::Enduser::UI::User;
use Net::Domain qw(hostdomain);
use CGI;
use CGI::Cookie;
use HTML::Template;

our  @ISA = ('PureMessage::Manager::Response');

sub new {
    my $proto = shift;
    my %params = @_;
    $params{'delay-headers'} = 1 unless defined $params{'delay-headers'};
    my $self = $proto->SUPER::new(%params);
    $self->{'content-type'} = 'text/html' unless $self->{'content-type'};
    $self->{'charset'} = 'utf-8' unless $self->{'charset'};
    if ($params{'no_image'}) {
	$self->{'no_image'} = 1;
    } else {
	$self->{'no_image'} = 0;
    }
    $self->set_handler('html_head', sub{$self->_html_head()});  
    if ($params{'titlebar-only'}) {
	$self->set_handler('content_start', sub{$self->_titlebar()});
    } else {
	$self->set_handler('content_start', sub{$self->_content_start()});
    }
    $self->set_handler('content_end', sub{$self->_content_end()});
    $self->set_handler('http_headers', sub{$self->_http_headers()});
    $self->set_handler('body_end', sub{});
    $self->{'nofeatures'} = 1 if $params{'nofeatures'};

    $self->{skin} = defined($enduser_ui_config{skin}) ? 
                            $enduser_ui_config{skin} : 'default';

    return $self;
}

sub _http_headers {
    my $self = shift;
    my @headers = ( '-type' => $self->{'content-type'},
		    '-charset' => $self->{'charset'},
		    '-Cache-Control' => 'no-cache',
		    '-Pragma' => 'no-cache',
		    '-Expires' => '-1',
		    );
    foreach my $header (keys %{$self->{'http-headers'}}) {
	next if $header =~ /content-type/;
	push(@headers,"-$header" => $self->{'http-headers'}->{$header});
    }
    my @cookies;
    foreach my $cookie_name (keys %{$self->{cookies}}) {
	my $cookie = $self->{cookies}->{$cookie_name};
        if ($cookie_name eq 'pmx_sid' and $cookie->value()) {
          $cookie->expires('+33d');
        }
        if ($cookie_name eq 'pmx_auth_email' and $cookie->value()) {
          $cookie->expires('+33d');
        }
	push(@cookies,$cookie);
    }
    push(@headers,-cookie => \@cookies);
    print $self->{REQUEST}->{cgi}->header(@headers);
}

# expects there to be either:
#  pmx_sid cookie set
#  auth_email, auth_password cgi variables

sub authorize_user {
    my $self = shift;
    my $cookie = $self->get_cookie('pmx_sid');
    my $sid;
    if ($cookie) {
	$sid = $cookie->value();
    }

    if($self->{REQUEST}->{query}->{pmx_sid}) {
       $sid = $self->{REQUEST}->{query}->{pmx_sid};
       $self->{user} = PureMessage::Enduser::UI::User->restore($sid);
       if ($self->{user}->valid_session) {
         $self->add_cookie('pmx_sid',$self->{user}->session_id(),'+33d');
       } 
       else {
         # the pmx_sid we retrieved is invalid
	 $self->force_logout();
         my %settings = $self->{user}->get_settings(); # for error
         $self->send;
	 exit;
       }
    } elsif($sid) {
	$self->{user} = PureMessage::Enduser::UI::User->restore($sid);
        if ($self->{user}->valid_session()) {
         $self->update_cookie('pmx_sid',$sid);
        } 
        else {
         # the pmx_sid we retrieved is invalid
	 $self->force_logout();
         my %settings = $self->{user}->get_settings(); # for error
         $self->send;
	 exit;
       }
    } else {
	my ($email,$pass) = ($self->{REQUEST}->{query}->{auth_email},$self->{REQUEST}->{query}->{auth_password});
	$self->{user} = $self->user( create => 1); # make a new user
	unless ($email && $pass) {
	    $self->force_logout();
            my %settings = $self->{user}->get_settings(); # for error
            $self->send;
	    exit;
	}
	if (! $self->{user}->authorize($email,$pass)) {
	    $self->force_logout(retry => 1);
            if ($self->{user} and my $x = $self->{user}->extended_error()) {
              print STDERR $x->{'error_authenticating_user'}->{errorstr} . "\n" if $x->{'error_authenticating_user'};
              $self->{user}->handle_error('error_authenticating_user');
            }
            # the default log unhandled errors will log for us
            $self->send;
	    exit;
	}
	$self->add_cookie('pmx_sid',$self->{user}->session_id(),'+33d');
	$self->add_cookie('pmx_auth_email', $email, '+33d');
    }
    return $self->{user};
}

sub logout_user {
  my $self = shift;
  $self->update_cookie('pmx_sid',0,"+0s"); # cause it to expire now!
  $self->update_cookie('pmx_auth_email',0,"+0s");
}

# get a user object, if create is true then if one isn't found make one
sub user {
  my ($self,%opts) = @_;
  if($opts{create} and !$self->{user}) {
    $self->{user} = PureMessage::Enduser::UI::User->new();
  } 
  return $self->{user};
}

sub add_cookie {
  my $self = shift;
  if (ref($_[0]) eq 'CGI::Cookie') {
    $self->update_cookie(@_);
  } else {
    my ($name,$value,$expire) = @_;
    $value = [$value] unless ref($value); # for somereason non array cookies aren't staying
    $expire = "+2d" unless $expire;
    my $cookie = new CGI::Cookie( -name=> $name,
                                  -value=>$value,
                                  -expires=>$expire,
                                );
    $self->update_cookie($cookie);
  }
}

sub update_cookie {
  my $self = shift;
  if (ref($_[0]) eq 'CGI::Cookie') {
    my $cookie = shift;
    $self->{cookies}->{$cookie->name} = $cookie;
  } else {
    my ($name,$value,$expire) = @_;
    $value = [$value] unless ref($value); # for somereason cookies that aren't arrays aren't sticking
    my $cookie = $self->{REQUEST}->{cookies}->{$name};
    return unless ref($cookie) eq 'CGI::Cookie';
    $cookie->value($value);
    $cookie->expires($expire) if $expire;
    $self->{cookies}->{$name} = $cookie;
    return 1;
  }
}
sub get_cookie_value {
  my $self = shift;
  my $cookie_name = shift;
  my $cookie = $self->get_cookie($cookie_name);
  return unless $cookie;
  my @value;
  @value = $cookie->value();
  return $value[0] unless wantarray;
  return @value;
}

sub get_cookie {
  my $self = shift;
  my $cookie_name = shift;
  return $self->{REQUEST}->{cookies}->{$cookie_name};
}

sub set_title {
    my $self = shift;
    $self->{content_start_data}->{title} = shift;
}

sub message {
   my $self = shift;
   $self->{content_start_data}->{message} = shift;
}


sub force_logout {
  my $self = shift; 
  my %options = @_;
  $self->clear_parts(); # remove any content that might already have been sent
  $self->logout_user(); 
  $self->set_handler('content_start', sub{$self->_titlebar()});
  my $authorize_tmpl = $self->new_tmpl('authorize.html');
  # set language name space
  PureMessage::Manager::Lang::set_namespace("EUUI/response");
  my $lang = $self->{user}->language();
  my $no_login =
      PureMessage::Enduser::Features::get_all->{'email.notification'} ?
      lang_text_utf8($lang, 'no_login') :
      '&nbsp;';
  $authorize_tmpl->param('invalid_auth' => lang_text_utf8($lang, 'invalid_auth'),
			 'login_instructions' => lang_text_utf8($lang, 'login_instructions'),
			 'action_email_login' => lang_text_utf8($lang, 'action_email_login'),
			 'action_password' => lang_text_utf8($lang, 'action_password'),
			 'button_login' => lang_text_utf8($lang, 'button_login'),
			 'no_login' => $no_login,
			 );
  $authorize_tmpl->param(retry => 1) if $options{retry};
  $self->add_part($authorize_tmpl);
  $self->{nofeatures} = 1;
}

sub redirect {
    my ($self, $url) = @_;
    my @cookies;
    foreach my $cookie_name (keys(%{$self->{cookies}})) {
	push (@cookies, $self->{cookies}->{$cookie_name});
    }
    print $self->{'REQUEST'}->{'cgi'}->redirect( -url => $url,
						 -cookie => \@cookies);
}



# overload this so rpc failures aren't silent.
sub send {
  my $self = shift;

  if($self->{user}) {
    
    # deal with errors first
    if($self->{user}->error()) {
       if(my $x = $self->{user}->extended_error()) {
          
          foreach my $error (keys %$x) {
              # deal with individual errors we know about
              if ($error eq 'rpc_invalid_session' or $error eq 'error_rpc_invalid_user_in_session') {
                # their cookie is broken so we need to fix it and allow
                # them to login again
                $self->force_logout();
                $self->raise_error($x->{$error}->{errorstr});
                last;  
              }elsif($error eq 'quarantine_query_too_soon') {
                $self->raise_error($x->{quarantine_query_too_soon}->{errorstr});
                $self->{user}->handle_error('quarantine_query_too_soon');
              }elsif($error eq 'unavailable') {
                $self->clear_parts();
                $self->set_handler('content_start', sub{$self->_titlebar()});
                $self->{nofeatures} = 1;
                my $unav_tmpl = $self->new_tmpl('unavailable.html');
                $self->add_part($unav_tmpl);
                my $msg = $x->{$error}->{errorstr} || $config{unavailable_banner};
                $unav_tmpl->param(msg => $msg);
                last;
              }else {
                 # not sure what happened so log it, for now assume we can continue
                 # to display what ever content was created by the cgi
                 my @p;
                 push(@p,"$_='$x->{$error}->{$_}'") foreach keys(%{$x->{$error}});
                 print STDERR "RPC Error: " . join(' ',@p) . "\n";
              }
          }

       } else {
         # not an extended error, log it
         print STDERR "RPC Error:" . $self->{user}->error() . "\n";
         
         # have a look and see if it's a fatal error we know about
         if ($self->{user}->error() =~ /^Host '.*' never returned EOF|^Net::HTTPS?::NB: connect: Connection refused/) {
             # Our connection to the RPC host timed out or it's not responding
             # to rpc requests
             $self->clear_parts();
             my $unav_tmpl = $self->new_tmpl('unavailable.html');
             $self->add_part($unav_tmpl);
             
             my $error_msg = lang_ntext_utf8('EUUI',$self->{user}->language(),'error_rpc_fatal_network') ne 'error_rpc_fatal_network'
                           ? lang_ntext_utf8('EUUI',$self->{user}->language(),'error_rpc_fatal_network')
                           : '';
             if($error_msg) {
                 my $info = $self->InfoBox(msg => $error_msg);
                 $unav_tmpl->param(error => $info->output());
             }
             $self->set_handler('content_start', sub{$self->_titlebar()});
             $self->{nofeatures} = 1;
             my $msg = $config{unavailable_banner};
             $unav_tmpl->param(msg => $msg);
         }
       }

    };
   
    if (!$self->{nofeatures}) {

        # this should be safe to do even with errors
        my %features = $self->{user}->get_features();
        $self->{features} = \%features;
           
        if (! %features) {
          my $err = $self->{user}->error() || '';
          carp("Problem getting features: $err");
        }
        
    } else {
      $self->{features} = {}; 
    }

  }# we don't have a user object 
  $self->SUPER::send(@_);
}

sub errors {
    my $self = shift;
    my $user = shift;
    my $prog_name = shift;
    my $comment = shift;
    my @known_errors = @_;
    $prog_name =~ s/.*\///g;
    my ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst) = localtime(time());
    my @months = ('Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec');
    my $date = "[" . $months[$mon] . " " . $mday . " " . ($year + 1900) . " " . sprintf("%02d", $hour) . ":" . sprintf("%02d", $min) . ":" . sprintf("%02d", $sec) . "]";
    return 0 unless my $err = $user->error();
    chomp $err;
    my $err_string = "Error: $err";
    my $handle_err = 0;
    my $fatal_err = 0;
    my $response_handled_error = 0;
    my %response_handled_errors = map { $_ => 1 } qw( error_rpc_invalid_user_in_session rpc_invalid_session quarantine_query_too_soon unavailable);
    if(my $xerr = $user->extended_error()) {
	if (ref($xerr) eq 'HASH') {
	    my $handle_xerr = 1;
	    foreach my $xerr_item (keys(%$xerr)) {
                # ignore errors handled by Response
		if ($response_handled_errors{$xerr_item}){
		    $response_handled_error =1;
                } else {
                    $handle_err = 1;
                    $handle_xerr = 1;
                    # known errors are not fatal
                    if (grep({$_ eq $xerr_item} @known_errors) == 0) {
                        $fatal_err = 1;
                    }
                }
		if (ref($xerr->{$xerr_item}) eq 'HASH') {
		    foreach my $sub_xerr_item (keys(%{$xerr->{$xerr_item}})) {
			$err_string .= ', Extended Error: ' . $xerr_item . "=>" . $sub_xerr_item . "=>" . $xerr->{$xerr_item}->{$sub_xerr_item};
			# raise errors if needed
			if (($handle_xerr) && ($sub_xerr_item eq 'errorstr')) {
                            if ($fatal_err) {
                                # set language name space
                                PureMessage::Manager::Lang::set_namespace("EUUI/response");
                                my $lang = $self->{user}->language();
                                $self->raise_error(lang_text_utf8($lang, 'error_contact_admin'));
                            } else {
                                $self->raise_error($xerr->{$xerr_item}->{$sub_xerr_item});
                            }
			}
		    }
                }
		if ($handle_xerr) {
		    $user->handle_error($xerr_item);
		    $handle_xerr = 0;
		}
	    }
        }
    }
    if ($handle_err) {
	$user->handle_error();
    }
    print STDERR $date . $prog_name . " - ERROR - " . $comment . " - " . $err_string . "\n";
    # In the case of a fatal error, send what we have and then stop.
    if ($fatal_err or $response_handled_error) {
        $self->send();
        exit 1;
    }
    return 1;
}

sub clear_parts {
  my $self = shift;
  $self->{parts} = [];
}

sub _tmpl {
    my $name = shift;
    return "$config{prefix}/lib/manager/HTTPD/tmpl/$name";
}

sub new_tmpl {
    my $self = shift;
    my $template_name = shift;
    my $new_template = HTML::Template->new(filename => _tmpl($template_name));
    return $new_template;
}

sub _html_head {
    my $self = shift;

    my $show_image = $self->{'no_image'} ? 0 : 1;
    my $html_head = $self->new_tmpl("_enduser_html_header.html");
    if (! defined $html_head) {
	carp "Could not instatiate template: _enduser_html_header.html";
    }
    $html_head->param('BACKGROUND_IMAGE' => $show_image,
                      'skin' => $self->{skin},
                        );

    print $html_head->output;
}

# set the navigation menu
# - when instatiating the EndUser class, one must pass a hash
#   to the new method
# - one of these hash keys should be called 'content_start_data'
# - this key should refer to a hash ref which contains two keys:
#   - 'title' which refers to a string equal to on of:
#     - 'title_block_messages'
#     - 'title_deleted_messages'
#     - 'title_approved_senders'
#     - 'title_blocked_senders'
#     - 'title_options'
#     - 'title_help'
#  - 'message' which refers to a string which will appear as a
#    confirmation in the navigation
#
sub _content_start {
    my $self = shift;
    my $content_start_data = $self->{'content_start_data'};
    my $title = $content_start_data->{'title'};
    my $message = $content_start_data->{'message'};

    # a little error checking in case a title is not given ... just keeps it from complaining
    if (defined $title) {
	if ($title !~ /^title/) {
	    if ($title !~ /blocked_messages|deleted_messages|approved_senders|blocked_senders|options/) {
		carp "Wrong title sent: $title";
	    }
	    $title = "title_" . $title;
	} else {
	    if ($title !~ /title_blocked_messages|title_deleted_messages|title_approved_senders|title_blocked_senders|title_options/) {
		carp "Wrong title sent: $title";
	    }
	}
    }
    
    my %features;
    %features = %{$self->{features}} if ref($self->{features}) eq 'HASH';
    my $options = $features{'option.spam_optout'} || $features{'option.digest_optout'} || $features{'option.vacation'} || $features{'language.preference'};
    my $body_attrib = $self->new_tmpl("_enduser_content_start.html");
    my %user = $self->{user}->get_user();
    my $language = $self->{user}->language();
    if (defined $title) {
	eval $body_attrib->param($title => 1);
    }
    # set language name space
    PureMessage::Manager::Lang::set_namespace("EUUI/response");
    my $lang = $self->{user}->language();
    my $lang_sophos;
    if (defined $lang) {
	if ($lang eq 'en') {
	    $lang_sophos = "www.sophos.com";
	} elsif ($lang eq 'fr') {
	    $lang_sophos = "www.sophos.fr";
	} elsif ($lang eq 'de') {
	    $lang_sophos = "www.sophos.de";
	} elsif ($lang eq 'es') {
	    $lang_sophos = "esp.sophos.com";
	} elsif ($lang eq 'it') {
	    $lang_sophos = "www.sophos.it";
	}
    } else {
	$lang_sophos = "www.sophos.com";
    }
    $body_attrib->param('main_title_blocked_messages' => lang_text_utf8($lang, 'main_title_blocked_messages'),
			'main_title_deleted_messages' => lang_text_utf8($lang, 'main_title_deleted_messages'),
			'main_title_approved_senders' => lang_text_utf8($lang, 'main_title_approved_senders'),
			'main_title_blocked_senders' => lang_text_utf8($lang, 'main_title_blocked_senders'),
			'main_title_options' => lang_text_utf8($lang, 'main_title_options'),
			'menu_blocked_messages' => lang_text_utf8($lang, 'menu_blocked_messages'),
			'menu_deleted_messages' => lang_text_utf8($lang, 'menu_deleted_messages'),
			'menu_approved_senders' => lang_text_utf8($lang, 'menu_approved_senders'),
			'menu_blocked_senders' => lang_text_utf8($lang, 'menu_blocked_senders'),
			'menu_options' => lang_text_utf8($lang, 'menu_options'),
			'menu_help' => lang_text_utf8($lang, 'menu_help'),
			'menu_logout' => lang_text_utf8($lang, 'menu_logout'),
			'quarantine_messages' => $features{'quarantine.messages'},
                        'user_info' => lang_text_utf8($lang, 'user_info'),
                        'email' => $user{email},
			'list_white' => $features{'list.white'},
			'list_black' => $features{'list.black'},
			'option_options' => $options,
			'left_nav_message' => $message,
			'lang' => $language,
			'lang_sophos' => $lang_sophos,
                         'skin' => $self->{skin},
			);
    print $body_attrib->output;
}

sub _content_end {
    my $self = shift;
    my $body_end = $self->new_tmpl("_enduser_content_end.html");
    print $body_end->output;
}

sub _titlebar {
    my $self = shift;
    my $titlebar = $self->new_tmpl("_titlebar.html");
    $titlebar->param('skin' => $self->{skin});
    print $titlebar->output;
}

1;
