package PureMessage::Enduser::UI::User::direct;

use Time::Zone;
use PureMessage::Config;
use PureMessage::Enduser::UI::Config qw(%config);
use PureMessage::Enduser::Config qw(%enduser_config);
use PureMessage::Manager::Lang qw(mtext text lang_ntext_utf8 lang_text_utf8);
use PureMessage::Enduser::Auth qw(Authorize);
use PureMessage::Enduser::Session;
use PureMessage::Enduser::Features;
use PureMessage::Logger;
use Data::Dumper;
use PureMessage::Resources::DBH qw(DBH);
use HTTP::Date qw(str2time time2iso);
use PureMessage::Enduser::User;
use base qw(PureMessage::QUser);

use Time::HiRes qw( usleep ualarm gettimeofday tv_interval );

use strict;
use warnings;

our $DEFAULT_LANGUAGE = $PureMessage::Config::config{language} || 'en';
our $USER_LIST;

our %TIMER;

our @ATTRIBUTES = qw( vacation_till 
                    spam_optout  
                    digest_optout
                    language    
                  );


=item new

Create and return a Remote::User object. The object will still need to
be authorized to be useable.

=cut

sub new {
  my $proto = shift;
  my $class = ref($proto) || $proto;
  my %self = @_;
  my $self = bless(\%self, $class);
  delete($self->{session_id});
  delete($self->{attribute});
  $self->{logger} = PureMessage::Logger->new(ident => "eui_direct");
  my @reasons = ref($enduser_config{allow}->{reasons}) eq 'ARRAY' ?
      @{$enduser_config{allow}->{reasons}} : ($enduser_config{allow}->{reasons});
  $self->{allowed_reasons} = \@reasons;
  return $self;
}


=item restore($session_id)

Create and return a user object that assumes that the passed session_id is a valid id.

=cut
 # this is a constructor, bootstrap this user from session


### TODO: XXX dpl this needs to be rewritten but just not yet....

sub restore {
  my ($proto,$session_id) =@_;
  my $class = ref($proto) || $proto;
  my %self = ();
  my $self = bless(\%self, $class);
  $self->{session_id} = $session_id;
  $self->{session} = $self->session();
  $self->{logger} = PureMessage::Logger->new(ident => "eui_direct");
  return $self unless $self->{session};
  $self->{email} = $self->{session}->param('email');
  
my @reasons = ref($enduser_config{allow}->{reasons}) eq 'ARRAY' ?
   @{$enduser_config{allow}->{reasons}} : ($enduser_config{allow}->{reasons});
$self->{allowed_reasons} = \@reasons;
  unless($self->{email}) { 
    $self->set_extended_error($self->_fextended($self->_fextended('error_rpc_invalid_user_in_session',$self->{email}))) ;
    $self->{logger}->log('INFO',"Invalid user in session $self->{session}: '$self->{email}'");
    return;
  }


  return $self;
}

=item authorize($email,$password)

Attempts to authenticate the user. If successful the return value is this users session_id
otherwise it is undef and C<error> is set.  If this object was created using the restore method, this user is assumed to be authorized,
this is probably not what you want. In order to confirm that this session is valid you
must attempt an operation with it or use the C<valid_session> method.

=cut


sub authorize {
  my ($self,$email,$password) = @_;
  return if $self->_unavailable();
  # TODO XXX this will do some odd stuff if we don't check that an existing
  # session has been marked by session->authorize

  # if we have a session_id we are going to assume it is valid! 
  return $self->{session_id} if $self->{session_id}; 
  
  $self->{logger} = PureMessage::Logger->new(ident => "eui_direct");

  unless($email) {
      $self->set_extended_error($self->_fextended('authorize_invalid_identity',$email)); 
      $self->{logger}->log('INFO',"Invalid identity for user '' in authentication.");
      return; 
  };
  
  $self->{logger} = PureMessage::Logger->new(ident => "eui_direct");

  my %info;
  eval {
      %info = Authorize($email,$password,logger => $self->{logger})
  };

  if (%info) { 

    unless( $self->valid_user( lc($info{email}) ) ) {
        
      $self->set_extended_error($self->_fextended('authorize_invalid_identity',$email)); 
      $self->{logger}->log("INFO", "user '$self->{email}' has an email address that doesn't match a valid user as defined by the list '$enduser_config{allow}->{users}'") if $self->{logger};
      return;
   
    }
    
    $self->{email} = lc( $info{email} );
    $self->{auth} = \%info;
    $self->{session} = $self->session(create => 1); 
    $self->{session_id} = $self->{session}->id(); 
    $self->{session}->authorize($self->{email});
    $self->{session}->flush();
  } else {
    return; # nope nope bad password
  }
  #warn "session_id is $self->{session_id}";
  return $self->{session_id}; 
}

#sub __ {
#  my ($self, $key, @info) = @_;
#  return ( $key => { errorstr => lang_text_utf8($self->language(), $key, @info) } );
#}

sub _fextended {
  my ($self,$lang_id,@info) = @_;
  return ( $lang_id => { errorstr => lang_ntext_utf8('PureMessage::RPC::Response::Enduser',$self->language(),$lang_id,@info) } );
#  return ( $lang_id => { errorstr => lang_ntext_utf8('PureMessage::RPC::Response::Enduser','en',$lang_id,@info) } );
#  return ( $lang_id => { errorstr => lang_text_utf8($self->language, $lang_id, @info) } );
}

# This sub is only for expanding the "attribute error" errors. We need
# a separate routine since in this case part of the error text is always
# in English, and part of it is translated - at least this how it is in
# 5.2 and earlier releases.
sub _fextended_attr_err {
  my ($self,$lang_id,@info) = @_;
#  return ( $lang_id => { errorstr => lang_ntext_utf8('PureMessage::RPC::Response::Enduser',$self->language(),$lang_id,@info) } );
  return ( $lang_id => { errorstr => lang_ntext_utf8('PureMessage::RPC::Response::Enduser','en',$lang_id,@info) . "<br>" . lang_text_utf8($self->language(),'err_problem_saving') } );
}

=item session

Returns the users last know session object or undef if none.

=cut

sub session {
  my $self = shift;
  my $session_id = $self->{session_id};
  my %options = @_;
  my $session;
  if ($session_id) {
    $session = PureMessage::Enduser::Session->restore($session_id);
  }
  
  unless($session) {
    if ($options{create}) {
      $session = PureMessage::Enduser::Session->new();
      return unless $session;
      $self->{session_id} = $session->id();
    }  else {
      delete($self->{session_id});
    }
  }
  return $session;
}

=item session_id 

Return the users session_id if any.

=cut

sub session_id {
  my $self = shift;
  return $self->{session_id} if $self->{session_id};
  return;
}

=item email_credentials($email_to)

Have the backend email an authorization email to the passed email address.

=cut

sub email_credentials {
  my $self = shift;
  my $email = shift || '';
  return if $self->_unavailable();

  # more email validation should be needed here
  $self->{logger}->log('INFO', mtext('email_credentials',$email));
  return 0 unless $email;
    $self->{logger}->log('INFO', "continued with invalid email");
  my $user = new PureMessage::Enduser::User($email,logger => $self->{logger});
  my $result;
  eval {
    $result = $user->send_notification();
  };

  if($@ || !$result) {
    $self->{logger}->log('ERROR', mtext('rpc_error_sending_notification_log',$email,$@));
    return 0;
  }

  return 1;
}

=item valid_session

Connects to the rpc backend and tests to see if the currently configured session_id is valid.
You will probably never need to call this directly as attempting an operation with an invalid
session_id will cause that call to failure and C<error> being set to a useful value.

=cut

sub valid_session{
  my $self = shift;
  return unless $self->{session_id};
  return 1;
}



=item valid_user($username)

Returns true if this user a user that is allowed to connect as defined
by the $config{allow}->{users} list.

=cut


sub valid_user {
  my $self = shift;
  my $user = shift;
  return unless $enduser_config{allow}->{users}; # by default allow none

  unless ($USER_LIST) {
    eval { $USER_LIST = $self->_get_resource($enduser_config{allow}->{users})->native; };
    unless ($USER_LIST) {
        $USER_LIST = PureMessage::List->new($enduser_config{allow}->{users});
    }
  }

  # TODO XXX this allowed users stuff should go away!!
  $self->set_extended_error($self->_fextended('user_invalid_access_list',$enduser_config{allow}->{users})),return unless $USER_LIST;
   
  return $USER_LIST->match($user); 
}


=item get_features

Returns as a hash the value of all the configured features for the
currently authorized user.  Features are settings like: quarantine.list
or quarantine.view that are used by the UI to autoconfigure it's self
to the admins settings.

=cut

sub get_features {
  my $self = shift;
  return if $self->_unavailable();
  return unless $self->{session_id}; # this call requires a session_id
  my $f;
  eval {
      $f = PureMessage::Enduser::Features->get_all(); 
  }; 

  if ($@) {
    $self->set_extended_error($self->_fextended('unavailable',errorstr=>$config{unavailable_banner})) ;
    $self->{logger}->log('INFO',"Error retrieving features.");
    return;
  }
  return %$f;
}


=item language

Attempts to determine the users language settings

=cut

sub language {
  my $self = shift;
  # XXXX TODO this is hacky as hell!
  return $DEFAULT_LANGUAGE unless $self->{session_id};
  my %v = $self->get_attribute('language');
  return $v{language};
}


sub email {
  my $self = shift;
  return $self->{email};
}


=item get_user()

Returns all the known information about this user

=cut

sub get_user {
  my $self = shift;
  return unless $self->{session_id}; # this call requires a session_id
  $self->{session} = $self->session() unless $self->{session};
  return unless $self->{session};
  #warn "get_user found all it needed, reading the session\n";
  my %info;
  $info{email} = $self->{session}->param('email');
  return %info;
}


=item get_attribute([$attribute])

Returns the value of the named attribute or all attributes as a hash if $attribute
is undef.

=cut


sub get_attribute {
  my $self = shift;
  return if $self->_unavailable();
  return () unless $self->{session_id}; # this call requires a session_id
  my @r =  $self->_get_attribute(@_);
  return @r;
} 



sub _get_resource {
   my $self = shift;
   my $resource = shift;
   my $res;
   
   eval {
       my $rm = $self->_resource_manager();
       
       eval {
         $res = $rm->find($resource);
       };
     
      if($@) { 
        $self->set_extended_error($self->_fextended('unavailable',errorstr=>$config{unavailable_banner})) ;
        $self->{logger}->log('INFO',"Failed to find the resource '$resource': $@");
        return;
      }
       
       unless($res) {
          eval {
            $res = $rm->add($resource);
          };
               
          if($@ or !$res) { 
            $self->set_extended_error($self->_fextended('unavailable',errorstr=>$config{unavailable_banner})) ;
            $self->{logger}->log('INFO',"Failed to find or add the resource '$resource': $@");
            return;
          }
       }

   
   };
   
   if(!$res or $@) {
    $self->set_extended_error($self->_fextended('unavailable',errorstr=>$config{unavailable_banner})) ;
    $self->{logger}->log('INFO',"Error initializing required resource '$resource': $@");
    return;
   } else {
     return $res;
   }
}


sub _resource_manager {
  my $self = shift;
  return $self->{resources_manager} if $self->{resources_manager};
  $self->{resources_manager} = PureMessage::Resources::Manager->new( writable => 1);
  return $self->{resources_manager};
}


sub _init_prefs {
  my $self = shift;
  return $self->{prefs} if $self->{prefs};
  my $prefs = $self->_get_resource('enduser-preferences');
  
  unless($prefs) {
    $self->set_extended_error($self->_fextended('unavailable',errorstr=>$config{unavailable_banner})) ;
    $self->{logger}->log('INFO',"Unable to load enduser preferences db");
    return;
  }
  
  $self->{prefs} = $prefs;
  return $prefs;
}



sub _get_attribute {
  my $self = shift;
  my %result;
  
  my @attributes = scalar(@_) ? @_ : @ATTRIBUTES;

  foreach my $an_attrib (@attributes) {
    my $val = $self->_get_single_attribute($an_attrib);
    $result{$an_attrib} = $val if defined($val);
  }
  return %result;

}


# to actual implement a new attribute you need not only to add it to this
# sub but add it's name into @ATTRIBUTES

sub _get_single_attribute {
  my ($self,$attribute) = @_;
 
  $self->_init_prefs();
  my $native = $self->{prefs}->native();
  
  if ($attribute eq 'vacation_till') {
     $self->{attribute}->{vacation_till} = $native->expire_delay_iso( $self->email() );
     return $self->{attribute}->{vacation_till};

  } elsif ($attribute eq 'language') {
     return $self->{attribute}->{language} ||= $native->language( $self->email()) || $DEFAULT_LANGUAGE; 
     
  } elsif ($attribute eq 'spam_optout') {
     return $self->{attribute}->{spam_optout} if $self->{attribute}->{spam_optout}; 
     $self->{attribute}->{spam_optout} = $self->_spam_optout();    
     return $self->{attribute}->{spam_optout};
     
  } elsif ($attribute eq 'digest_optout') {
     
     return $self->{attribute}->{digest_optout} if $self->{attribute}->{digest_optout}; 
     $self->{attribute}->{digest_optout} = $self->_digest_optout();
     return $self->{attribute}->{digest_optout};

  } else {

     $self->set_extended_error($self->_fextended('unavailable',errorstr=>$config{unavailable_banner})) ;
     $self->{logger}->log('INFO',"Request for unknown attribute '$attribute' from '" . $self->email() . "'");
     return;
  }
}

=item set_attribute([$attribute => $value, .. ])

Sets the value of the passed attributes.

=cut


sub set_attribute {
  my $self = shift;
  my %attributes = @_;
  return if $self->_unavailable();
  delete($self->{attribute}); # remove our cached stuff
  
  return unless $self->{session_id}; # this call requires a session_id
  
  foreach my $attr (keys %attributes) {
    $self->_set_single_attribute($attr,$attributes{$attr});
  }
  return %attributes;
}

sub _set_single_attribute {
  my ($self,$attribute,$value) = @_;
  my $r;
  
  if ($attribute eq 'vacation_till') {

     $self->_init_prefs();
     $r= $self->_set_expire_delay_iso_value($value);
  } elsif ($attribute eq 'language') {
      
     $self->_init_prefs();
     $r= $self->{prefs}->native->language( $self->email(),$value );
     
  } elsif ($attribute eq 'spam_optout') {
      
     $r =  $self->_spam_optout($value);    
     
  } elsif ($attribute eq 'digest_optout') {
     
     $r = $self->_digest_optout($value);

  } else {

     $self->set_extended_error($self->_fextended('unavailable',errorstr=>$config{unavailable_banner})) ;
     $self->{logger}->log('INFO',"Request for unknown attribute '$attribute' from '" . $self->email() . "'");
     return;
  }

  delete($self->{attribute}->{$attribute});
  return $r;
}


sub _set_expire_delay_iso_value {
  my ($self,$value) = @_;

  # we are expecting an iso formatted date: YYYY-MM-DD 
  unless( $value =~ /^\d{4}-[0-1]?\d-[0-3]?\d$/ ) {
     $self->set_extended_error($self->_fextended('unavailable',errorstr=>$config{unavailable_banner})) ;
     $self->{logger}->log('INFO',"Date '$value' is not a valid iso date.");
     return;
  }
  
  my $time_value = str2time($value);
    
  if ($time_value > time + $enduser_config{limit}->{vacation_expiry}) {

    my $days = int($enduser_config{limit}->{vacation_expiry}/(60*60*24));

    my $error_msg = "$value exceeds (current date + $days days)";
    $self->set_extended_error($self->_fextended_attr_err('set_attribute_error', 'hold days', $error_msg));
    $self->{logger}->log('INFO',"Attempt to set excessive hold days for user '" . $self->email() . "', " . $error_msg);
    return; 
  } else {

      $self->{prefs}->native->expire_delay_iso( $self->email(),$value );
      # no longer valid so delete it
      delete($self->{prefs});

  }

}

=item boolean_value

  Utility sub (NOT METHOD) that can be used to determine if a value is true if interpreted
  as a boolean value.

=cut

sub boolean_value {
  my $val = shift;
  return 1 if $val =~ /^1|yes|true$/i; 
  return 0;
}

# **** XXX TODO NOTE ****
# this is a bit heavy handed but we are *NOT* going to check and see if
# this user or his inverse is in the list we are just going to remove
# both expressions that *might* be there and add the one we want.
# Whatever! It's a huge improvement over enumerating the entire list just
# to see if this items exists in it.  Same goes for ALl list opt-in/out
# operations


sub _spam_optout {
  my $self = shift;

  $self->{anti_spam_optouts} = $self->_get_resource('anti-spam-optouts')
    unless $self->{anti_spam_optouts};

  if (scalar(@_)) {
    my $ld = $self->{anti_spam_optouts}->native->data();
    
    # our operations on this list are going to invalidate it
    delete($self->{anti_spam_optouts});
    
    # you might be asking your self why we are removing the exact thing
    # we are going to add... well that's so we don't get dups of course :) 
    $ld->remove_items($self->email(),'!' . $self->email() );
   
    my $bool = boolean_value(shift);

    if($bool) {
       $ld->add_items($self->email());
    } else {
       $ld->add_items('!' . $self->email());
    }
    $ld->commit();
    return $bool; 
  } else {
    my $r =  $self->{anti_spam_optouts}->native->match( $self->email() ); 
    return $r;
  }
  
}

sub _digest_optout {
  my $self = shift;
  
  $self->{digest_users} = $self->_get_resource('digest-users')
    unless $self->{digest_users};
  
  if (scalar(@_)) {
    # XXX TODO this is cut and paste code, unify with the above sub
    
    my $ld = $self->{digest_users}->native->data();
    
    # our operations on this list are going to invalidate it
    delete($self->{digest_users});
    
    # you might be asking your self why we are removing the exact thing
    # we are going to add... well that's so we don't get dups of course :) 
    $ld->remove_items($self->email(),'!' . $self->email() );
   
    my $bool = boolean_value(shift);
    #warn "setting _digest_optout to $bool\n";
    
    if(! $bool) {
       $ld->add_items($self->email());
    } else {
       $ld->add_items('!' . $self->email());
    }

    $ld->commit();
    return $bool; 
 
  } else {
    my $r = ! $self->{digest_users}->native->match( $self->email() ); 
    return $r;
  }
}


=item get_lists()

Return a hash of available lists. Keys are the list_id's with the values
being the description that describes this list.

=cut


my %EDITABLE_LISTS =  %{$enduser_config{lists}};

sub get_lists {
  my $self = shift;
  return if $self->_unavailable();
  return unless $self->{session_id}; # this call requires a session_id
  return %EDITABLE_LISTS;
}

=item set_items($list_id,\@items)

Set the contents of list C<$list_id> to be @items;

=cut

sub set_items {
  my ($self,$public_name,$items) = @_;
  return unless $self->{session_id}; # this call requires a session_id
  return if $self->_unavailable();

  if (scalar(@$items) > $enduser_config{limit}->{max_list_items}) {
    $self->{logger}->log('WARNING',"List $public_name: item count exceeds maximum; list will not be updated");
    return;
  }
 
  my $list_id = $EDITABLE_LISTS{$public_name};

  unless($list_id) { 
    $self->set_extended_error($self->_fextended('unavailable',errorstr=>$config{unavailable_banner})) ;
    $self->{logger}->log('INFO',"Invalid list '$public_name'.");
    return;
  }

  my $mlist = $self->_get_resource($list_id)->native();
  my $list;
  
  if (UNIVERSAL::isa($mlist,'PureMessage::MultiList')) {
      $list = $mlist->find_list(key => $self->email, edit_mode => 1);
  } else {
      $list = $mlist;
  }

  my $data = $list->data();
  $data->clear();
  $data->commit();
  $data->add_items(@$items);
  $data->commit();
  return 1;
}


=item get_items($list_id)

Return a reference to an array of items in C<$list_id>.

=cut


sub get_items {
  my ($self,$list_name) = @_;
  return unless $self->{session_id}; # this call requires a session_id
  return if $self->_unavailable();
 
  my $list_id = $EDITABLE_LISTS{$list_name};

  unless($list_id) { 
    $self->set_extended_error($self->_fextended('unavailable',errorstr=>$config{unavailable_banner})) ;
    $self->{logger}->log('INFO',"Invalid list '$list_name'.");
    return;
  }

  my $mlist = $self->_get_resource($list_id)->native();
  my $list;
  
  if (UNIVERSAL::isa($mlist,'PureMessage::MultiList')) {
      $list = $mlist->find_list(key => $self->email, edit_mode => 1);
  } else {
      $list = $mlist;
  }

  unless($list) { 
    $self->set_extended_error($self->_fextended('unavailable',errorstr=>$config{unavailable_banner})) ;
    $self->{logger}->log('INFO',"Invalid list '$list_name'.");
    return;
  }

  my $data = $list->data();
  
  my @items;
  $data->enumerate_items( sub { my $item = shift; push(@items,$item)});
  #warn "get_items returned " . join(', ',@items);
  return \@items;
}

=item get_settings()

Return a hash of non-updatable config settings.

=cut


sub get_settings {
  my ($self) = @_;
  my %info;
  return if $self->_unavailable();
  return if $self->_unavailable();
  $info{max_vacation} = $enduser_config{limit}->{vacation_expiry}; # in seconds
  $info{query_wait} = $enduser_config{limit}->{query_wait}; # in seconds
  $info{quarantine_expiry_days} = $enduser_config{quarantine_expiry_days} || -1; # in days
  return %info;
}


# check to see if we are allowed to load or if this is disabled. returns true
# if euwi is currently configured as unavailable
sub _unavailable {
  my $self = shift;

  unless($enduser_config{enabled}) {
    # admin has turned off access to this interface
    $self->set_extended_error($self->_fextended('unavailable',errorstr=>$config{unavailable_banner})) ;
    $self->{logger}->log('INFO',"Interface is currently disabled.");
    return 1;
  }
  return;
}


=item error()

Returns the last error if any.  Error handling is a bit complicated but the basic process is that
error should return a string error message if there is an error and extended_error should return
a hash structure of error data.


=cut

sub error {
  my $self = shift;
  return $self->{error} if $self->{error};
  return;
}


=item extended_error() 

Returns a hash reference of extended errors if
there were any returned by an rpc call. 
 
  if ($user->error()) {
    if(my $x = $user->extended_error()) {
      foreach my $e ( keys %x) {
        print "Error $e occurred: " 
            . $x->{$e}->{errorstr} 
            . "\n";
      }
    }
  }

The format of an extended error is: 
  { manager_lang_string => {
                                errorstr => 'human displayable string';
                                # additional info for error debugging
                           }
    # more lang_strings if there are more errors
  }


The errors for these lang strings are mostly within the C<PureMessage::RPC::Response::Enduser>
or C<PureMessage::Enduser::User> lang namespace.  Be aware that in some instances the errorstr 
above WILL BE DISPLAYED DIRECTLY TO USERS so they must be formatted in an appropriately 
professional manner.  

To get the appropriate lang entry we need to use the ntext method thusly:

  use PureMessage::Manager::Lang qw( ntext lang_ntext_utf8 ); # use utf is localized
  my $errorstr = ntext('PureMessage::Enduser::User','query_to_soon',$allowedfrequency,$remainingseconds); 
  # note that the error will be throw as: C<quarantine_query_too_soon> that
  # resides in the C<PureMessage::RPC::Response::Enduser> namespace

  # confused yet?
  

The currently throw errors are (although I've tried this list is NOT exhaustive):

  rpc_invalid_session
    # were given a bad sid in the cookie or header

  error_rpc_invalid_user_in_session
    # although we could find the session passed in the cookie the format of it
    # is either something we don't understand or the user contained within the
    # session isn't allowed to do the requested operation

  quarantine_query_too_soon
    # we are rate limiting quarantine queries (which doesn't make much sense
    # with the db.pm backend) but the user has to come back later to requery
    # perhaps we should limit the number of transactions in a sliding window
    # instead so that compulsive refreshers aren't lagging the box

  unavailable
    # the admin has temporarily disabled access to the euwi.  Note that it
    # should be possible for the admin to do this and still have the ui
    # display this error instead of a "database is missing/broken" error


=cut


sub extended_error {
  my $self = shift;
  return $self->{extended_error} if $self->{extended_error};
  return;
}


# internal method that can be used to throw an extended error
sub set_extended_error {
  my ($self,%errors) = @_; 
  %errors = (unavailable => { errorstr => 'Operation currently unavailable; please contact your administrator.'})
      unless scalar %errors;
  foreach my $k (keys %errors) {
    $self->{extended_error}->{$k} = $errors{$k};
  }
  $self->update_error_from_extended_error();
}


# this is an internal method that makes sure that the string set into error is
# always set when there is an extended error.  There should never be a need to
# call this method directly 
sub update_error_from_extended_error {
    my $self = shift;
    $self->{error} = join(' ',map{$self->{extended_error}->{$_}->{errorstr}} keys %{$self->{extended_error}});
}

# delete an error that was dealt with, if there args they are assumed to be the
# names of the specific keys to delete
sub handle_error {
  my ($self,$name) = @_;
  if ($name) {
    delete($self->{extended_error}->{$name});
    $self->update_error_from_extended_error();
  } else {
    delete($self->{extended_error});
    delete($self->{error});
  }
}

# The following few subroutines are simple wrappers for their counterparts
# from the base class. The only thing we're adding here is raising errors.
# note that we're not logging the errors here - this is expected to have
# already been done.

=item get_qview
sub get_qview {
    my ($self, @view) = @_;
    my %view_count;
    eval {
        %view_count = $self->SUPER::get_qview(@view);
    }
    if ($@) {
        $self->set_extended_error(
            $self->_fextended('unavailable',
                               errorstr=>$config{unavailable_banner}
                             )
        );
    }
    return wantarray ? %view_count : $self->{views}->{$view[0]};
}
=cut

1;
