package PureMessage::Enduser::UI::User::rpc;

use PureMessage::RPC::ClientV2;
use PureMessage::Config;
use PureMessage::Enduser::UI::Config qw(%config);
use PureMessage::Manager::Lang qw(mtext text );
use PureMessage::RPC::NamedDataEncoder;
use Data::Dumper;
use strict;
use warnings;
our $DEFAULT_LANGUAGE = $PureMessage::Config::config{language} || 'en';

=head1 NAME

PureMessage::Enduser::UI::User - Interface to RPC user operations

=head1 SYNOPSIS

  use PureMessage::Enduser::UI::User;

  my $user;
  
  #### first time we've seen this user 
  $user = new PureMessage::Enduser::UI::User();
  
  if (!$user->authorize($users_email,$password)) {
    # unable to authorize this user.. wrong password?
    die $user->error();
  }
  my %features = $user->get_features();
  
  my $session_id = $user->session_id(); # to store in a cookie  
   
  ... 

  #### reconstruct this user from a session_id
  $user = PureMessage::Enduser::UI::User->restore($session_id);
 
  my %info = $user->init_qview('visible');
    
  my @rows = $user->get_qrows('visible',0,$info{visible}); # get all rows
  
  ...  

  
=head1 DESCRIPTION

This object can be used to communicate with the configured RPC backend 
services.  All RPC calls are handled transparently as if they were
local.

=cut

=item new

Create and return a Remote::User object. The object will still need to
be authorized to be useable.

=cut

sub new {
  my $proto = shift;
  my $class = ref($proto) || $proto;
  my %self = @_;
  my $self = bless(\%self, $class);
  $self->_init_rpc();
  return $self;
}

=item restore($session_id)

Create and return a user object that assumes that the passed session_id is a valid id.

=cut

sub restore {
  my ($proto,$session_id) =@_;
  my $class = ref($proto) || $proto;
  my %self = ();
  my $self = bless(\%self, $class);
  $self->{session_id} = $session_id;
  $self->_init_rpc();
  $self->_init_lang();
  return $self;
}


# setup the RPC::Client object, and do some sanity checking on config
# settings.

sub _init_rpc {
  my $self = shift;
  my %rpc_data = ( host  => $config{RPC}->{host},
                   proto => uc($config{RPC}->{proto}), #RPC expects HTTPS not https
                   port  => $config{RPC}->{port},
                  );
                 
  $self->{rpc_host} = $config{RPC}->{host} unless $self->{rpc_host};
  $self->set_error(mtext('rpc_host_is_invalid',$self->{rpc_host})) unless $self->{rpc_host};
  warn mtext('rpc_host_is_invalid',$self->{rpc_host}) unless $self->{rpc_host};

  $self->{rpc_client} =  PureMessage::RPC::ClientV2->new({$self->{rpc_host} => \%rpc_data},timeout => $config{RPC}->{timeout});
  $self->set_error(mtext('rpc_connect_failure',$rpc_data{host},$rpc_data{proto},$rpc_data{prot})) unless $self->{rpc_client};
  warn mtext('rpc_connect_failure',$rpc_data{host},$rpc_data{proto},$rpc_data{prot}) unless $self->{rpc_client};

  $self->{en} = new PureMessage::RPC::NamedDataEncoder();
}

=item authorize($email,$password)

Attempts to authenticate the user. If successful the return value is this users session_id
otherwise it is undef and C<error> is set.  If this object was created using the restore method, this user is assumed to be authorized,
this is probably not what you want. In order to confirm that this session is valid you
must attempt an operation with it or use the C<valid_session> method.

=cut

sub authorize {
  my ($self,$email,$password) = @_;

  # if we have a session_id we are going to assume it is valid! 
  return $self->{session_id} if $self->{session_id}; 

  if (my @results = $self->_rpc('auth',$email,$password)) {
    # TODO I'd like to extend the Enduser.auth rpc call to return more data
    if (ref($results[0]) eq 'HASH' and my $session_id = $results[0]->{session}) {
      $self->{session_id} = $session_id;
      return $session_id if $self->_init_lang();
      return; 
    } else {
      # that's odd
      $self->set_error(mtext('invalid_rpc_data_returned','Enduser.auth'));
      return;
    }
  } else {
    # error will be set for us by the rpc call
    return;
  }
}

=item session_id 

Return the users session_id if any.

=cut

sub session_id {
  my $self = shift;
  return $self->{session_id} if $self->{session_id};
  return;
}

=item valid_session

Connects to the rpc backend and tests to see if the currently configured session_id is valid.
You will probably never need to call this directly as attempting an operation with an invalid
session_id will cause that call to failure and C<error> being set to a useful value.

=cut

sub valid_session{
  my $self = shift;
  return unless $self->{session_id};
  $self->_init_lang();
  return 1 if $self->_init_lang();
  return;
}

=item email_credentials($email_to)

Have the backend email an authorization email to the passed email address.

=cut

sub email_credentials {
  my ($self,$email_to) = @_;
  return unless $email_to;
  my @results = $self->_rpc('email_credentials',$email_to);
  return unless defined($results[0]); # we got nothing back
  return $results[0];
}

=item get_features

Returns as a hash the value of all the configured features for the
currently authorized user.  Features are settings like: quarantine.list
or quarantine.view that are used by the UI to autoconfigure it's self
to the admins settings.

=cut

sub get_features {
  my $self = shift;
  return unless $self->{session_id}; # this call requires a session_id
  
  if (my @results = $self->_rpc('get_features',$self->{session_id})) {
    unless(!scalar(@results) or !defined($results[0]) or ref($results[0]) eq 'HASH') {
      # we got some bad data from the RPC backend
      $self->set_error(mtext('invalid_rpc_data_returned','Enduser.get_features'));
    };
    return %{$results[0]};
    
  } else {
    # error will be set for us by the rpc call
    return;
  }
}


=item get_attribute([$attribute])

Returns the value of the named attribute or all attributes as a hash if $attribute
is undef.

=cut


sub get_attribute {
  my $self = shift;
  my @attribute = @_;
  return unless $self->{session_id}; # this call requires a session_id
  
  if (my @results = $self->_rpc('get_attribute',$self->{session_id},@attribute)) {
    unless(ref($results[0]) eq 'HASH') {
      # we got some bad data from the RPC backend
      $self->set_error(mtext('invalid_rpc_data_returned','Enduser.get_attribute'));
    };
    return %{$results[0]};
    
  } else {
    # error will be set for us by the rpc call
    return;
  }
}


=item set_attribute([$attribute => $value, .. ])

Sets the value of the passed attributes.

=cut


sub set_attribute {
  my $self = shift;
  my %attributes = @_;
  return unless $self->{session_id}; # this call requires a session_id
  
  if (my @results = $self->_rpc('set_attribute',$self->{session_id},%attributes)) {
    unless(ref($results[0]) eq 'HASH') {
      # we got some bad data from the RPC backend
      $self->set_error(mtext('invalid_rpc_data_returned','Enduser.set_attribute'));
    };
    $self->{language} = $attributes{language} if exists($attributes{language});
    return %{$results[0]};
    
  } else {
    # error will be set for us by the rpc call
    return;
  }
}

=item get_user()

Returns all the known information about this user

=cut

sub get_user {
  my $self = shift;
  return unless $self->{session_id}; # this call requires a session_id

  if (my @results = $self->_rpc('get_user', $self->{session_id})) {
    unless(ref($results[0]) eq 'HASH') {
      # we got some bad data from the RPC backend
      $self->set_error(mtext('invalid_rpc_data_returned','Enduser.get_user'));
    }
    return %{$results[0]};
  } else {
    # error will be set for us by the rpc call
    return;
  }
}

=item language

Attempts to connect to the rpc backend and determine the users language settings

=cut

sub language {
    my $self = shift;
    $self->_init_lang() unless $self->{language};
    return $self->{language};
}

sub _init_lang {
    my $self = shift;
    my %attribute = $self->get_attribute();
    if (%attribute) {
        $self->{language} = $attribute{language};
        return $attribute{language};
    } 
    else {
        $self->{language} = $DEFAULT_LANGUAGE;
        return;
    }
}


=item init_qview([$qview, .. ])

Initialize the remote quarantine view. If an array context returns a hash of view
names and the count of visible items. Otherwise returns a count of visible items.
If no view name is passed then all views are initialized. Currently configured views
are: visible, approve, delete, all.

=cut

sub init_qview {
  my $self = shift;
  my @qviews = @_; 
  return unless $self->{session_id}; # this call requires a session_id
  
  if (my @results = $self->_rpc('init_qview',$self->{session_id},@qviews)) {
    unless(ref($results[0]) eq 'HASH') {
      # we got some bad data from the RPC backend
      $self->set_error(mtext('invalid_rpc_data_returned','Enduser.init_qview'));
    };
    my (%views,$count);
    foreach my $row (@results) {
      $views{$row->{view}} = $row->{row_count};
      $count += $row->{row_count};
    }

    return %views if wantarray;
    return $count;
    
  } else {
    # error will be set for us by the rpc call
    return;
  }
}
=item get_qview([$qview, .. ])

If an array context returns a hash of view names and the count of visible items. 
Otherwise returns a count of visible items.  If no view name is passed then 
all views used in the UI are returned. Currently configured views are: 
visible, delete.

=cut

sub get_qview {
  my $self = shift;
  my @qviews = @_; 
  return unless $self->{session_id}; # this call requires a session_id
  
  if (my @results = $self->_rpc('get_qview',$self->{session_id},@qviews)) {
    unless(ref($results[0]) eq 'HASH') {
      # we got some bad data from the RPC backend
      $self->set_error(mtext('invalid_rpc_data_returned','Enduser.get_qview'));
    };
    my (%views,$count);
    $count = 0;
    foreach my $row (@results) {
      $views{$row->{view}} = $row->{row_count};
      $count += $row->{row_count} if grep { $_ eq $row->{view} } @qviews;
    }

    return %views if wantarray;
    return $count;
    
  } else {
    # error will be set for us by the rpc call
    return;
  }
}



=item get_qrows($qview,$start,$end [,$sort[,$descending]])

Returns an array of hashes. One row each for each message in the
quarantine view from $start to $end. Keys for each message include:
qid m_enduser c_reason_csv h_from h_to m_date h_subject m_spam_probability

An optional field_name to order the returned rows by can be passed. If no
sort arg is passed then rows will be returned in whatever the last sort
ordere was.

=cut

sub get_qrows {
  my $self = shift;
  my ($qview,$start,$end,$sort,$desc) = @_;
  return unless $self->{session_id}; # this call requires a session_id
  my @sort;
  if ($sort) {
    $sort = "-$sort" if $desc; # rpc expects desc fields to just have a - prepended
    push(@sort,$sort,'m_date:tzt');
    push(@sort,'m_date:tzt') unless grep /m_date:tzt/, @sort; # we default to probablity, date
  }
  if (my @results = $self->_rpc('get_qrows',$self->{session_id},$qview,$start,$end,@sort)) {
    unless(ref($results[0]) eq 'HASH') {
      # we didn't get any data back from the call
      return ();
    };
    return @results if wantarray;
    return $#results;
    
  } else {
    # error will be set for us by the rpc call
    return;
  }
}


=item get_msg_body($msg_id)

Gets the full body information for the passed msg_id. This will return a 
hash. The keys of this hash are the same as those returned from get_msg_meta
with the addition of a C<body> key where the msg body may be found.

=cut

sub get_msg_body {
  my $self = shift;
  my $id = shift;
  return unless $self->{session_id}; # this call requires a session_id
  return unless $id;
  
  if (my @results = $self->_rpc('getmsg_body',$self->{session_id},$id)) {
    unless(ref($results[0]) eq 'HASH') {
      # we got some bad data from the RPC backend
      $self->set_error(mtext('invalid_rpc_data_returned','Enduser.getmsg_meta'));
      return;
    };
    return $results[0];
    
  } else {
    # error will be set for us by the rpc call if there is one
    return;
  }
}



=item get_msg_meta(@msg_ids)

Gets the meta information for the passed msg_ids. This is the same data as is
returned by the get_qrow method.

=cut

sub get_msg_meta {
  my $self = shift;
  my @ids = @_;
  return unless $self->{session_id}; # this call requires a session_id
  
  if (my @results = $self->_rpc('getmsg_meta',$self->{session_id},@ids)) {
    unless(ref($results[0]) eq 'HASH') {
      # we got some bad data from the RPC backend
      $self->set_error(mtext('invalid_rpc_data_returned','Enduser.getmsg_meta'));
      return;
    };
    return @results;
    
  } else {
    # error will be set for us by the rpc call if there is one
    return;
  }
}


=item approve_msg(@msg_ids)

Approve and release the passed msg_ids; Returns a hash of msg_ids and the
result of the operation on them. If in a scalar context returns the count
of items that succeded.

=cut


sub approve_msg {
  my $self = shift;
  my @msg_ids = @_; 
  return unless $self->{session_id}; # this call requires a session_id
  
  if (my @results = $self->_rpc('approve_msg',$self->{session_id},@msg_ids)) {
    unless(ref($results[0]) eq 'HASH') {
      # we got some bad data from the RPC backend
      $self->set_error(mtext('invalid_rpc_data_returned','Enduser.approve_msg'));
      return;
    };
    my %result;
    my $count = 0;
    # result has a row for each qid passed, keys in that row are the qid and
    # value is true or false 
    foreach my $row (@results) {
      foreach my $key (%$row) {
        $result{$key} = $row->{$key};
        $count++ if$result{$key}; # count how many were true
      }
    }
    return %result if wantarray;
    return $count;
    
  } else {
    # error will be set for us by the rpc call
    return;
  }
}


=item delete_msg(@msg_ids)

Delete the passed msg_ids; Returns a hash of msg_ids and the
result of the operation on them. If in a scalar context returns the count
of items that succeded.

=cut


sub delete_msg {
  my $self = shift;
  my @msg_ids = @_; 
  return unless $self->{session_id}; # this call requires a session_id
  
  if (my @results = $self->_rpc('delete_msg',$self->{session_id},@msg_ids)) {
    unless(ref($results[0]) eq 'HASH') {
      # we got some bad data from the RPC backend
      $self->set_error(mtext('invalid_rpc_data_returned','Enduser.delete_msg'));
    };
    my %result;
    my $count = 0;
    # result has a row for each qid passed, keys in that row are the qid and
    # value is true or false 
    foreach my $row (@results) {
      foreach my $key (keys %$row) {
        $result{$key} = $row->{$key};
        $count++ if$result{$key}; # count how many were true
      }
    }
    return %result if wantarray;
    return $count;
    
  } else {
    # error will be set for us by the rpc call
    return;
  }
}

=item undelete_msg(@msg_ids)

Undelete the passed msg_ids; Returns a hash of msg_ids and the
result of the operation on them. If in a scalar context returns the count
of items that succeded.

=cut


sub undelete_msg {
  my $self = shift;
  my @msg_ids = @_; 
  return unless $self->{session_id}; # this call requires a session_id
  
  if (my @results = $self->_rpc('undelete_msg',$self->{session_id},@msg_ids)) {
    unless(ref($results[0]) eq 'HASH') {
      # we got some bad data from the RPC backend
      $self->set_error(mtext('invalid_rpc_data_returned','Enduser.undelete_msg'));
    };
    my %result;
    my $count = 0;
    # result has a row for each qid passed, keys in that row are the qid and
    # value is true or false 
    foreach my $row (@results) {
      foreach my $key (keys %$row) {
        $result{$key} = $row->{$key};
        $count++ if$result{$key}; # count how many were true
      }
    }
    return %result if wantarray;
    return $count;
    
  } else {
    # error will be set for us by the rpc call
    return;
  }
}


=item get_lists()

Return a hash of available lists. Keys are the list_id's with the values
being the description that describes this list.

=cut


sub get_lists {
  my $self = shift;
  return unless $self->{session_id}; # this call requires a session_id
  
  if (my @results = $self->_rpc('get_lists',$self->{session_id})) {
    unless(ref($results[0]) eq 'HASH') {
      # we got some bad data from the RPC backend
      $self->set_error(mtext('invalid_rpc_data_returned','Enduser.get_lists'));
    };
    return %{$results[0]};
    
  } else {
    # error will be set for us by the rpc call
    return;
  }
}

=item get_settings()

Return a hash of non-updatable config settings.

=cut


sub get_settings {
  my ($self) = @_;
   
  if (my @results = $self->_rpc('get_settings')) {
    unless(ref($results[0]) eq 'HASH') {
      # we got some bad data from the RPC backend
      $self->set_error(mtext('invalid_rpc_data_returned','Enduser.get_settings'));
      return;
    };
    return %{$results[0]};
  } else {
    return ();
  }
}


=item get_items($list_id)

Return a reference to an array of items in C<$list_id>.

=cut


sub get_items {
  my $self = shift;
  my $id = shift;
  return unless $self->{session_id}; # this call requires a session_id
  
  if (my @results = $self->_rpc('get_items',$self->{session_id},$id)) {
    unless(!scalar(@results) or !defined($results[0]) or ref($results[0]) eq 'ARRAY' ) {
      # we got some bad data from the RPC backend
      $self->set_error(mtext('invalid_rpc_data_returned','Enduser.get_items'));
    };
    return $results[0];
    
  } else {
    # error will be set for us by the rpc call
    return;
  }
}


=item set_items($list_id,\@items)

Set the contents of list C<$list_id> to be @items;

=cut


sub set_items {
  my ($self,$id,$items) = @_;
  return unless $self->{session_id}; # this call requires a session_id
  
  if (my @results = $self->_rpc('set_items',$self->{session_id},$id,@$items)) {
    return $results[0]; 
  } else {
    # error will be set for us by the rpc call
    return;
  }
}

=item error()

Returns the last error if any

=cut

sub error {
  my $self = shift;
  return $self->{error} if $self->{error};
  return;
}


=item extended_error() 

Returns a hash reference of extended errors if
there were any returned by an rpc call. 
 
  if ($user->error()) {
    if(my $x = $user->extended_error()) {
      foreach my $e ( keys %x) {
        print "Error $e occurred: " 
            . $x->{$e}->{errorstr} 
            . "\n";
      }
    }
  }
=cut

sub extended_error {
  my $self = shift;
  return $self->{extended_error} if $self->{extended_error};
  return;
}

#internal method used to set error's
sub set_error {
  my ($self,$error) = @_;
  $self->{rpc_error} = undef;
  $self->{error} = $error;
}

#internal method used to set rpc errors
sub set_rpc_error {
  my $self = shift;
  $self->{rpc_error} = $#_ >= 0 ? 1 : 0; # if we have a string set the error
  $self->set_error(@_);
}

# delete an error that was dealt with, if there args they are assumed to be the
# names of the specific keys to delete
sub handle_error {
  my ($self,$name) = @_;
  if ($name) {
    delete($self->{extended_error}->{$name});
    delete($self->{rpc_error}) unless scalar(keys(%{$self->{extended_error}}));
    $self->{error} = join(' ',map{$self->{extended_error}->{$_}->{errorstr}} keys %{$self->{extended_error}});
  } else {
    delete($self->{extended_error});
    delete($self->{rpc_error});
    delete($self->{error});
  }
}

my %CACHEABLE = map { $_ => 1 } qw( get_features get_qview get_settings get_items get_attribute); 
sub cache_results {
  my ($self,$results) = @_;
  return unless ref($results) eq 'HASH';
  foreach my $key (keys %$results) {
    if($CACHEABLE{$key}) {
      my @to_cache = $self->massage_results($key,$results->{$key});
      $self->{cache}->{$key} = \@to_cache;
    }
  }
}

sub have_cache {
  my ($self,$key) = @_;
  return $self->{cache}->{$key} if defined($self->{cache}->{$key});
  return;
}

my %ARRAY = ( get_items => 1 );
sub massage_results {
  my ($self,$call,$r) = @_;
  my @result;
  if ($ARRAY{$call}) {
    # force the result to be an array
    $r = [$r] unless ref($r) eq 'ARRAY';
    @result = ($r);
  } else {
    @result = ref($r) eq 'ARRAY' ? @$r : ($r);
  }
  # print "call: $call\n";
  # print Dumper(\@result);
  return @result;
}

# internal method using to make the actual rpc calls
sub _rpc {
  my ($self,$call,@args) = @_;
  return if $self->error();
  if (my $cache = $self->have_cache($call)) {
    return @$cache;
  }
# TODO thi should be bulletproof, eval and catch exceptions
  my $client = $self->{rpc_client};
  my $rpc_call = $call =~ /Enduser/ ? $call : "EnduserUI.$call";
  
  $client->call($rpc_call,@args);
  $client->wait();
  if($client->error()) {
     $self->set_rpc_error($client->error());
     return;
  } else {
    # we have rpc data
    my @data = $client->data($self->{rpc_host});
    my ($results,$error) = $self->{en}->parse_response_data(@data);
    if ($results) {
        $self->cache_results($results); # even if there was an error we should inspect the return for data
        if ( scalar( keys(%$error) ) ) {
           $self->{extended_error} = $error;
           $self->set_rpc_error(join(' ',map{$error->{$_}->{errorstr}} keys %$error));
           return;
        } else {
          # we got data, make sure it's in the right format
          if(exists($results->{$call})) {
            my $r = $results->{$call};
            return $self->massage_results($call,$r);
          } else {
            return; # we got no data back
          }
        }
    } else {
      return @data;
    }
  }

}

=item set_items($list_id,\@items)

Set the contents of list C<$list_id> to be @items;

=cut

=head1 SEE ALSO

L<PureMessage::RPC::Response::Enduser>, L<PureMessage::Enduser::User>

=cut

1;

